﻿//
//                  BATCH FILE SYNTAX HIGHLIGHTER
//                          AddIn for POIDE
//                         By LD Blake, 2012
//

#define UNICODE
#define _UNICODE
#define WIN32_LEAN_AND_MEAN
#define WIN32_DEFAULT_LIBS

#include <windows.h>
#include <wchar.h>
#include <wctype.h>
#include <stdlib.h>
#include <addin.h>

// globalized variables
HINSTANCE hInst = NULL;
HWND      hMain = NULL;

///////////////////////////////////////////////////////////////////////////////
// Syntax Highlighting
//

// Flags for Cookie.
#define FLAG_COLORED       0x01  // set for coloured text
#define FLAG_SIMPLE        0x02  // inside simple command
#define FLAG_COMPLEX       0x04  // inside complex command
#define FLAG_NEWLINE       0x08  // first visible character
#define FLAG_CONTINUED     0x10  // line is continued

// colour setup
#define COLOR_KEYWORD   ADDIN_COLOR_KEYWORD
#define COLOR_MODIFIER  ADDIN_COLOR_FUNCTION
#define COLOR_CMDFLAG   ADDIN_COLOR_PREPROCESSOR
#define COLOR_VARIABLE  ADDIN_COLOR_NUMBER
#define COLOR_STRING    ADDIN_COLOR_STRING
#define COLOR_TEXT      ADDIN_COLOR_TEXT
#define COLOR_COMMENT   ADDIN_COLOR_COMMENT
#define COLOR_SYMBOL    ADDIN_COLOR_OPERATOR

// note: Keywords and Modifiers must be alpha sorted!
// Keywords go to either end of line, | or &
#define SIZE_SIMPLE 40
const PWCHAR Simple[] = 
    { L"ASSOC", L"BREAK", L"CALL", L"CD", L"CHDIR", L"CLS", L"COLOR", L"COPY", 
      L"DATE", L"DEL", L"DIR", L"ECHO", L"ENDLOCAL", L"ERASE", L"EXIT", 
      L"FTYPE", L"GOTO", L"MD", L"MKDIR", L"MKLINK", L"MOVE", L"PATH", 
      L"PAUSE", L"POPD", L"PROMPT", L"PUSHD", L"RD", L"REM", L"REN", L"RENAME", 
      L"RMDIR", L"SET", L"SHIFT", L"START", L"TIME", L"TITLE", L"TYPE", L"VER", 
      L"VERIFY", L"VOL" }; 

// complex keywords, may contain other keywords
#define SIZE_COMPLEX 6
const PWCHAR Complex[] =
    { L"DO", L"ELSE", L"FOR", L"IF", L"IN", L"SETLOCAL"}; 

// Modifieres, highlighted only inside complex commands
#define SIZE_MODIFIERS 21
const PWCHAR Modifiers[] =
    { L"DEFINED", L"DELIMS", L"DISABLEDELAYEDEXPANSION", L"DISABLEEXTENSIONS",
      L"ENABLEDALAYEDEXPANSION", L"ENABLEEXTENSIONS", L"EOL", L"EQU", L"ERRORLEVEL", 
      L"EXIST", L"GEQ", L"GTR", L"LEQ", L"LSS", L"NEQ", L"NOT", L"OFF", L"ON", 
      L"SKIP", L"TOKENS", L"USEBACKQ"   };

// characters before simple commands
const WCHAR CmdLead[] = {L" \t|&@("};
// characters that end simple commands
const WCHAR CmdBreak[] = {L")|&\n"};

// command line flags
const WCHAR FlagLead[] = {L"-/"};
const WCHAR FlagBreak[] = {L" :\n"};

// variables
//always start with %
const WCHAR VarBreak[] = {L" %\n"};

// stuff we highlight all the time
const WCHAR Symbols[] = {L"~^|&()<>@"};



// Check wordlist
BOOL IsKeyword (const PWCHAR List[], INT Size, PWSTR word)
  {
    INT hi = Size - 1;      // high end for binary search
    INT lo = 0;             // low end for binary search
    INT try;                // value being tried
    INT tst;                // result of test
    // binary search
    while (hi >= lo)
      {
        try = (hi + lo) / 2;
        tst = _wcsicmp(List[try],word);
        if (tst == 0)       // found!
          return 1;
        else if (tst < 0)   // look higher
          lo = try + 1;
        else 
          hi = try - 1;     // look lower
      }
    return 0;
  }



// helper for setting colors
void inline SetColor (ADDIN_PARSE_POINT Point[], PINT Points, INT Idx, INT Color)
  {
    Point[(*Points)].iChar = Idx;
    Point[(*Points)].iColor = Color;
    *Points += 1;
    return;
  }    



// POIDE color parser callback 
static USHORT CALLBACK BatParser(USHORT Cookie, PCTSTR Text, INT Len, ADDIN_PARSE_POINT Point[], PINT Points)
  { 
    INT offset  = 1;      // offset for keywords
    INT idx     = -1;     // text index 
    PWCHAR word;          // word to check

    // function is called with no points???
    if (Point == NULL || Len == 0)
      return Cookie;

    // space for word checks
    word = calloc(Len + 1,sizeof(WCHAR)); 
 
    // make sure no parse points defined
    *Points = 0;

    // mark a new line
    Cookie = FLAG_NEWLINE;

    // character loop
    while (idx < Len)    
      {
        // word jump through each line
        idx += offset;
 
        // fast bypass whitespace
        if (iswspace(Text[idx]))
          { 
            offset = 1;
            continue;
          }        

        // terminate simple command
        if (Cookie & FLAG_SIMPLE)
          if (wcschr (CmdBreak, Text[idx]))
            Cookie &= ~FLAG_SIMPLE;

        // : at beginning of line
        if ((Text[idx] == L':') && (Cookie & FLAG_NEWLINE))
          {
            if (Text[idx + 1] == L':')  // whole line comment
              SetColor (Point, Points, idx, COLOR_COMMENT);
            else                        // label
              SetColor (Point, Points, idx, COLOR_MODIFIER);
  
            Cookie &= ~FLAG_NEWLINE;
            Cookie |= FLAG_COLORED;
            offset = Len - idx;
            continue;
          }

        // special characters
        if (wcschr(Symbols,Text[idx]))
         {
            Cookie &= ~FLAG_NEWLINE;
            Cookie |= FLAG_COLORED;
            SetColor(Point, Points, idx, COLOR_SYMBOL);
            offset = 1;
            continue;
          }

        // command line flags
        if (!(Cookie & FLAG_SIMPLE))
          if (wcschr(FlagLead, Text[idx]))
            { 
              Cookie &= ~FLAG_NEWLINE;
              Cookie |= FLAG_COLORED;
              SetColor(Point, Points, idx, COLOR_CMDFLAG);

              // search for end flag
              for (offset = 1; offset < (Len - idx); offset++)
                if (wcschr(FlagBreak, Text[idx + offset]))
                  break;
              offset++;  // need to be one past end
              continue;
            }

        // variables
        if (Text[idx] == L'%')
          {
            Cookie &= ~FLAG_NEWLINE;
            Cookie |= FLAG_COLORED;
            SetColor(Point, Points, idx, COLOR_VARIABLE);

            // search for end flag
            for (offset = 1; offset < (Len - idx); offset++)
              if (wcschr(VarBreak, Text[idx + offset]))
                break;
            offset++; // point past breaker
            continue;
          }  

        // quoted strings
        if (! (Cookie & FLAG_SIMPLE))
          if (Text[idx] == L'"')
            {
              Cookie &= ~FLAG_NEWLINE;
              Cookie |= FLAG_COLORED;
              SetColor(Point, Points, idx, COLOR_STRING);

              // search for end flag
              for (offset = 1; offset < (Len - idx); offset++)
                if (Text[idx + offset] == L'"')
                  break;
              offset++; // point past breaker
              continue;
            }  

        // now we start checking keywords
        for (offset = 0; offset < (Len - idx); offset++)
          {
            word[offset] = Text[idx + offset];
            if (iswspace (Text[idx + offset]))
              break;
          }
        word[offset] = 0;
                          
        // simple commands
        if (! (Cookie & FLAG_SIMPLE))
          if (IsKeyword (Simple, SIZE_SIMPLE, word))
            if ((idx == 0) || wcschr (CmdLead, Text[idx - 1]))
              {
                Cookie &= ~FLAG_NEWLINE;
                Cookie |= FLAG_COLORED | FLAG_SIMPLE;
                SetColor (Point, Points, idx, COLOR_KEYWORD);
                continue;
              }
                
        // complex commands
        if (! (Cookie & FLAG_SIMPLE))
          if (IsKeyword (Complex, SIZE_COMPLEX, word))
            if ((idx == 0) || wcschr (CmdLead, Text[idx - 1]))
              {
                Cookie &= ~FLAG_NEWLINE;
                Cookie |= FLAG_COLORED | FLAG_COMPLEX;
                SetColor (Point, Points, idx, COLOR_KEYWORD);
                continue;
              }  

        // modifiers
        if (Cookie & FLAG_COMPLEX)         
          if (IsKeyword (Modifiers, SIZE_MODIFIERS, word))
            if ((idx == 0) || wcschr (CmdLead, Text[idx - 1]))
              {
                Cookie &= ~FLAG_NEWLINE;
                Cookie |= FLAG_COLORED;
                SetColor (Point, Points, idx, COLOR_MODIFIER);
                continue;
              }
        
        // first visible black character
        if (Cookie & FLAG_NEWLINE)
          Cookie &= ~FLAG_NEWLINE;

        // nothing special, reset to text
        if (Cookie & FLAG_COLORED)
          {
            Cookie &= ~FLAG_COLORED;
            SetColor(Point, Points, idx, COLOR_TEXT);
          }
      
        offset = 1;
      }
    free(word);
    return Cookie;
  }

///////////////////////////////////////////////////////////////////////////////
// AddIn Functions
//

// initialize the addin
BOOL InitAddIn(HWND Win)
  { 
    ADDIN_ADD_FILE_TYPE af = {0};
    hMain = Win;

    // Define a new file type in the IDE (.bat).
    af.cbSize = sizeof(af);
    af.pszDescription = L"Batch file";
    af.pszExtension = L"bat";  
    af.pfnParser = BatParser;
    af.pszShells = NULL;

    return AddIn_AddFileType(Win, &af);
  }



// main addin procedure
ADDINAPI BOOL WINAPI AddInMain(HWND Win, ADDIN_EVENT Event)
  { 
    switch (Event)
      { 
        case AIE_APP_CREATE :
          return InitAddIn (Win); 
        default:
            return 1;  
      } 
  }



// dll main entry point
BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved)
  { 
    hInst = hDLL;
    return TRUE;  
  }
